@@ -1,6 +1,6 @@
 /************* tabbson C++ Program Source Code File (.CPP) *************/
-/* PROGRAM NAME: tabbson     Version 1.0                               */
-/*  (C) Copyright to the author Olivier BERTRAND          2020         */
+/* PROGRAM NAME: tabbson     Version 1.2                               */
+/*  (C) Copyright to the author Olivier BERTRAND          2020 - 2021  */
 /*  This program are the BSON class DB execution routines.             */
 /***********************************************************************/
 
@@ -53,6 +53,7 @@ USETEMP UseTemp(void);
 bool    JsonAllPath(void);
 int     GetDefaultDepth(void);
 char   *GetJsonNull(void);
+bool    Stringified(PCSZ, char*);
 
 /***********************************************************************/
 /* BSONColumns: construct the result blocks containing the description */
@@ -158,8 +159,9 @@ BSONDISC::BSONDISC(PGLOBAL g, uint* lg)
   bp = NULL;
   row = NULL;
   sep = NULL;
+  strfy = NULL;
   i = n = bf = ncol = lvl = sz = limit = 0;
-  all = strfy = false;
+  all = false;
 } // end of BSONDISC constructor
 
 int BSONDISC::GetColumns(PGLOBAL g, PCSZ db, PCSZ dsn, PTOS topt)
@@ -172,8 +174,8 @@ int BSONDISC::GetColumns(PGLOBAL g, PCSZ db, PCSZ dsn, PTOS topt)
   lvl = GetIntegerTableOption(g, topt, "Depth", lvl);
   sep = GetStringTableOption(g, topt, "Separator", ".");
   sz = GetIntegerTableOption(g, topt, "Jsize", 1024);
-  limit = GetIntegerTableOption(g, topt, "Limit", 10);
-  strfy = GetBooleanTableOption(g, topt, "Stringify", false);
+  limit = GetIntegerTableOption(g, topt, "Limit", 50);
+  strfy = GetStringTableOption(g, topt, "Stringify", NULL);
 
   /*********************************************************************/
   /*  Open the input file.                                             */
@@ -186,10 +188,17 @@ int BSONDISC::GetColumns(PGLOBAL g, PCSZ db, PCSZ dsn, PTOS topt)
 #endif   // ZIP_SUPPORT
   tdp->Fn = GetStringTableOption(g, topt, "Filename", NULL);
 
+  if (!tdp->Fn && topt->http)
+    tdp->Fn = GetStringTableOption(g, topt, "Subtype", NULL);
+
   if (!(tdp->Database = SetPath(g, db)))
     return 0;
 
-  tdp->Objname = GetStringTableOption(g, topt, "Object", NULL);
+  if ((tdp->Objname = GetStringTableOption(g, topt, "Object", NULL))) {
+    if (*tdp->Objname == '$') tdp->Objname++;
+    if (*tdp->Objname == '.') tdp->Objname++;
+  } // endif Objname
+
   tdp->Base = GetIntegerTableOption(g, topt, "Base", 0) ? 1 : 0;
   tdp->Pretty = GetIntegerTableOption(g, topt, "Pretty", 2);
   tdp->Xcol = GetStringTableOption(g, topt, "Expand", NULL);
@@ -199,7 +208,8 @@ int BSONDISC::GetColumns(PGLOBAL g, PCSZ db, PCSZ dsn, PTOS topt)
   if (!tdp->Fn && !tdp->Uri) {
     strcpy(g->Message, MSG(MISSING_FNAME));
     return 0;
-  } // endif Fn
+  } else
+    topt->subtype = NULL;
 
   if (tdp->Fn) {
     //  We used the file name relative to recorded datapath
@@ -213,8 +223,7 @@ int BSONDISC::GetColumns(PGLOBAL g, PCSZ db, PCSZ dsn, PTOS topt)
 
   if (tdp->Uri) {
 #if defined(JAVA_SUPPORT) || defined(CMGO_SUPPORT)
-    tdp->Collname = GetStringTableOption(g, topt, "Name", NULL);
-    tdp->Collname = GetStringTableOption(g, topt, "Tabname", tdp->Collname);
+    tdp->Collname = GetStringTableOption(g, topt, "Tabname", NULL);
     tdp->Schema = GetStringTableOption(g, topt, "Dbname", "test");
     tdp->Options = (PSZ)GetStringTableOption(g, topt, "Colist", "all");
     tdp->Pipe = GetBooleanTableOption(g, topt, "Pipeline", false);
@@ -224,7 +233,7 @@ int BSONDISC::GetColumns(PGLOBAL g, PCSZ db, PCSZ dsn, PTOS topt)
       (tdp->Version == 2) ? "Mongo2Interface" : "Mongo3Interface");
     tdp->Pretty = 0;
 #else   // !MONGO_SUPPORT
-    sprintf(g->Message, MSG(NO_FEAT_SUPPORT), "MONGO");
+    snprintf(g->Message, sizeof(g->Message), MSG(NO_FEAT_SUPPORT), "MONGO");
     return 0;
 #endif  // !MONGO_SUPPORT
   } // endif Uri
@@ -236,7 +245,7 @@ int BSONDISC::GetColumns(PGLOBAL g, PCSZ db, PCSZ dsn, PTOS topt)
 #if defined(ZIP_SUPPORT)
       tjsp = new(g) TDBBSON(g, tdp, new(g) UNZFAM(tdp));
 #else   // !ZIP_SUPPORT
-      sprintf(g->Message, MSG(NO_FEAT_SUPPORT), "ZIP");
+      snprintf(g->Message, sizeof(g->Message), MSG(NO_FEAT_SUPPORT), "ZIP");
       return 0;
 #endif  // !ZIP_SUPPORT
     } else
@@ -252,7 +261,7 @@ int BSONDISC::GetColumns(PGLOBAL g, PCSZ db, PCSZ dsn, PTOS topt)
   } else {
     if (!((tdp->Lrecl = GetIntegerTableOption(g, topt, "Lrecl", 0)))) {
       if (!mgo) {
-        sprintf(g->Message, "LRECL must be specified for pretty=%d", tdp->Pretty);
+        snprintf(g->Message, sizeof(g->Message), "LRECL must be specified for pretty=%d", tdp->Pretty);
         return 0;
       } else
         tdp->Lrecl = 8192;       // Should be enough
@@ -267,7 +276,7 @@ int BSONDISC::GetColumns(PGLOBAL g, PCSZ db, PCSZ dsn, PTOS topt)
 #if defined(ZIP_SUPPORT)
       tjnp = new(g)TDBBSN(g, tdp, new(g) UNZFAM(tdp));
 #else   // !ZIP_SUPPORT
-      sprintf(g->Message, MSG(NO_FEAT_SUPPORT), "ZIP");
+      snprintf(g->Message, sizeof(g->Message), MSG(NO_FEAT_SUPPORT), "ZIP");
       return NULL;
 #endif  // !ZIP_SUPPORT
     } else if (tdp->Uri) {
@@ -275,14 +284,14 @@ int BSONDISC::GetColumns(PGLOBAL g, PCSZ db, PCSZ dsn, PTOS topt)
 #if defined(CMGO_SUPPORT)
         tjnp = new(g) TDBBSN(g, tdp, new(g) CMGFAM(tdp));
 #else
-        sprintf(g->Message, "Mongo %s Driver not available", "C");
+        snprintf(g->Message, sizeof(g->Message), "Mongo %s Driver not available", "C");
         return 0;
 #endif
       } else if (tdp->Driver && toupper(*tdp->Driver) == 'J') {
 #if defined(JAVA_SUPPORT)
         tjnp = new(g) TDBBSN(g, tdp, new(g) JMGFAM(tdp));
 #else
-        sprintf(g->Message, "Mongo %s Driver not available", "Java");
+        snprintf(g->Message, sizeof(g->Message), "Mongo %s Driver not available", "Java");
         return 0;
 #endif
       } else {             // Driver not specified
@@ -291,7 +300,7 @@ int BSONDISC::GetColumns(PGLOBAL g, PCSZ db, PCSZ dsn, PTOS topt)
 #elif defined(JAVA_SUPPORT)
         tjnp = new(g) TDBBSN(g, tdp, new(g) JMGFAM(tdp));
 #else
-        sprintf(g->Message, MSG(NO_FEAT_SUPPORT), "MONGO");
+        snprintf(g->Message, sizeof(g->Message), MSG(NO_FEAT_SUPPORT), "MONGO");
         return 0;
 #endif
       } // endif Driver
@@ -367,7 +376,7 @@ int BSONDISC::GetColumns(PGLOBAL g, PCSZ db, PCSZ dsn, PTOS topt)
       } // endswitch ReadDB
 
     } else
-      jsp = bp->GetArrayValue(bdp, i);
+      jsp = bp->GetNext(jsp);
 
     if (!(row = (jsp) ? bp->GetObject(jsp) : NULL))
       break;
@@ -428,7 +437,7 @@ bool BSONDISC::Find(PGLOBAL g, PBVAL jvp, PCSZ key, int j)
     jcol.Type = TYPE_UNKNOWN;
     jcol.Len = jcol.Scale = 0;
     jcol.Cbn = true;
-  } else  if (j < lvl) {
+  } else  if (j < lvl && !Stringified(strfy, colname)) {
     if (!fmt[bf])
       strcat(fmt, colname);
 
@@ -468,7 +477,7 @@ bool BSONDISC::Find(PGLOBAL g, PBVAL jvp, PCSZ key, int j)
         n = sizeof(fmt) - (strlen(fmt) + 1);
 
         if (!tdp->Xcol || stricmp(tdp->Xcol, key)) {
-          sprintf(buf, "%d", k);
+          snprintf(buf, sizeof(buf), "%d", k);
 
           if (tdp->Uri) {
             strncat(strncat(fmt, sep, n), buf, n - strlen(sep));
@@ -494,12 +503,12 @@ bool BSONDISC::Find(PGLOBAL g, PBVAL jvp, PCSZ key, int j)
 
       return false;
     default:
-      sprintf(g->Message, "Logical error after %s", fmt);
+      snprintf(g->Message, sizeof(g->Message), "Logical error after %s", fmt);
       return true;
     } // endswitch Type
 
   } else if (lvl >= 0) {
-    if (strfy) {
+    if (Stringified(strfy, colname)) {
       if (!fmt[bf])
         strcat(fmt, colname);
 
@@ -599,33 +608,51 @@ void BSONDISC::AddColumn(PGLOBAL g)
 /***********************************************************************/
 PBVAL BTUTIL::FindRow(PGLOBAL g)
 {
-  char *p, *objpath;
+  char *p, *objpath = PlugDup(g, Tp->Objname);
+  char *sep = (char*)(Tp->Sep == ':' ? ":[" : ".[");
+  bool  bp = false, b = false;
   PBVAL jsp = Tp->Row;
   PBVAL val = NULL;
 
-  for (objpath = PlugDup(g, Tp->Objname); jsp && objpath; objpath = p) {
-    if ((p = strchr(objpath, Tp->Sep)))
+  for (; jsp && objpath; objpath = p, bp = b) {
+    if ((p = strpbrk(objpath + 1, sep))) {
+      b = (*p == '[');
       *p++ = 0;
+    } // endif p
 
-    if (*objpath != '[' && !IsNum(objpath)) { // objpass is a key
+    if (!bp && *objpath != '[' && !IsNum(objpath)) { // objpass is a key
       val = (jsp->Type == TYPE_JOB) ?
         GetKeyValue(jsp, objpath) : NULL;
     } else {
-      if (*objpath == '[') {
-        if (objpath[strlen(objpath) - 1] == ']')
-          objpath++;
-        else
+      if (bp || *objpath == '[') {                   // Old style
+        if (objpath[strlen(objpath) - 1] != ']') {
+          snprintf(g->Message, sizeof(g->Message), "Invalid Table path %s", Tp->Objname);
           return NULL;
-      } // endif [
+        } else if (!bp)
+          objpath++;
+
+      } // endif bp
 
       val = (jsp->Type == TYPE_JAR) ?
-        GetArrayValue(GetArray(jsp), atoi(objpath) - Tp->B) : NULL;
+        GetArrayValue(jsp, atoi(objpath) - Tp->B) : NULL;
     } // endif objpath
 
       //  jsp = (val) ? val->GetJson() : NULL;
     jsp = val;
   } // endfor objpath
 
+  if (jsp && jsp->Type != TYPE_JOB) {
+    if (jsp->Type == TYPE_JAR) {
+      jsp = GetArrayValue(jsp, Tp->B);
+
+      if (jsp->Type != TYPE_JOB)
+        jsp = NULL;
+
+    } else
+      jsp = NULL;
+
+  } // endif Type
+
   return jsp;
 } // end of FindRow
 
@@ -649,17 +676,22 @@ PBVAL BTUTIL::MakeTopTree(PGLOBAL g, int type)
   if (Tp->Objname) {
     if (!Tp->Row) {
       // Parse and allocate Objpath item(s)
-      char* p;
-      char *objpath = PlugDup(g, Tp->Objname);
+      char *p, *objpath = PlugDup(g, Tp->Objname);
+      char *sep = (char*)(Tp->Sep == ':' ? ":[" : ".[");
       int   i;
+      bool  bp = false, b = false;
       PBVAL objp = NULL;
       PBVAL arp = NULL;
 
-      for (; objpath; objpath = p) {
-        if ((p = strchr(objpath, Tp->Sep)))
+      for (; objpath; objpath = p, bp = b) {
+        if ((p = strpbrk(objpath + 1, sep))) {
+          b = (*p == '[');
           *p++ = 0;
+        } // endif p
 
-        if (*objpath != '[' && !IsNum(objpath)) {
+
+        if (!bp && *objpath != '[' && !IsNum(objpath)) {
+          // objpass is a key
           objp = NewVal(TYPE_JOB);
 
           if (!top)
@@ -671,15 +703,15 @@ PBVAL BTUTIL::MakeTopTree(PGLOBAL g, int type)
           val = NewVal();
           SetKeyValue(objp, MOF(val), objpath);
         } else {
-          if (*objpath == '[') {
+          if (bp || *objpath == '[') {
             // Old style
             if (objpath[strlen(objpath) - 1] != ']') {
-              sprintf(g->Message, "Invalid Table path %s", Tp->Objname);
+              snprintf(g->Message, sizeof(g->Message), "Invalid Table path %s", Tp->Objname);
               return NULL;
-            } else
+            } else if (!bp)
               objpath++;
 
-          } // endif objpath
+          } // endif bp
 
           if (!top)
             top = NewVal(TYPE_JAR);
@@ -731,7 +763,6 @@ void BCUTIL::SetJsonValue(PGLOBAL g, PVAL vp, PBVAL jvp)
     case TYPE_FLOAT:
       switch (vp->GetType()) {
         case TYPE_STRING:
-        case TYPE_DATE:
         case TYPE_DECIM:
           vp->SetValue_psz(GetString(jvp));
           break;
@@ -749,9 +780,25 @@ void BCUTIL::SetJsonValue(PGLOBAL g, PVAL vp, PBVAL jvp)
           if (jvp->Type == TYPE_DBL || jvp->Type == TYPE_FLOAT)
             vp->SetPrec(jvp->Nd);
 
+          break;
+        case TYPE_DATE:
+          if (jvp->Type == TYPE_STRG) {
+            PSZ dat = GetString(jvp);
+
+            if (!IsNum(dat)) {
+              if (!((DTVAL*)vp)->IsFormatted())
+                ((DTVAL*)vp)->SetFormat(g, "YYYY-MM-DDThh:mm:ssZ", 20, 0);
+
+              vp->SetValue_psz(dat);
+            } else
+              vp->SetValue(atoi(dat));
+
+          } else
+            vp->SetValue(GetInteger(jvp));
+
           break;
         default:
-          sprintf(G->Message, "Unsupported column type %d", vp->GetType());
+          snprintf(G->Message, sizeof(G->Message), "Unsupported column type %d", vp->GetType());
           throw 888;
       } // endswitch Type
 
@@ -871,7 +918,7 @@ PBVAL BCUTIL::GetRowValue(PGLOBAL g, PBVAL row, int i)
       bvp = row;
       break;
     default:
-      sprintf(g->Message, "Invalid row JSON type %d", row->Type);
+      snprintf(g->Message, sizeof(g->Message), "Invalid row JSON type %d", row->Type);
       bvp = NULL;
     } // endswitch Type
 
@@ -881,7 +928,7 @@ PBVAL BCUTIL::GetRowValue(PGLOBAL g, PBVAL row, int i)
   } // endfor i
 
   return bvp;
-} // end of GetColumnValue
+} // end of GetRowValue
 
 /***********************************************************************/
 /*  GetColumnValue:                                                    */
@@ -1075,7 +1122,7 @@ PBVAL BCUTIL::GetRow(PGLOBAL g)
       val = row;
       break;
     default:
-      sprintf(g->Message, "Invalid row JSON type %d", row->Type);
+      snprintf(g->Message, sizeof(g->Message), "Invalid row JSON type %d", row->Type);
       val = NULL;
     } // endswitch Type
 
@@ -1143,10 +1190,15 @@ bool BSONDEF::DefineAM(PGLOBAL g, LPCSTR am, int poff)
   G = g;
   Schema = GetStringCatInfo(g, "DBname", Schema);
   Jmode = (JMODE)GetIntCatInfo("Jmode", MODE_OBJECT);
-  Objname = GetStringCatInfo(g, "Object", NULL);
+
+  if ((Objname = GetStringCatInfo(g, "Object", NULL))) {
+    if (*Objname == '$') Objname++;
+    if (*Objname == '.') Objname++;
+  } // endif Objname
+
   Xcol = GetStringCatInfo(g, "Expand", NULL);
   Pretty = GetIntCatInfo("Pretty", 2);
-  Limit = GetIntCatInfo("Limit", 10);
+  Limit = GetIntCatInfo("Limit", 50);
   Base = GetIntCatInfo("Base", 0) ? 1 : 0;
   Sep = *GetStringCatInfo(g, "Separator", ".");
   Accept = GetBoolCatInfo("Accept", false);
@@ -1157,7 +1209,7 @@ bool BSONDEF::DefineAM(PGLOBAL g, LPCSTR am, int poff)
     Collname = GetStringCatInfo(g, "Name",
       (Catfunc & (FNC_TABLE | FNC_COL)) ? NULL : Name);
     Collname = GetStringCatInfo(g, "Tabname", Collname);
-    Options = GetStringCatInfo(g, "Colist", NULL);
+    Options = GetStringCatInfo(g, "Colist", Xcol ? "all" : NULL);
     Filter = GetStringCatInfo(g, "Filter", NULL);
     Pipe = GetBoolCatInfo("Pipeline", false);
     Driver = GetStringCatInfo(g, "Driver", NULL);
@@ -1170,7 +1222,7 @@ bool BSONDEF::DefineAM(PGLOBAL g, LPCSTR am, int poff)
       Wrapname = GetStringCatInfo(g, "Wrapper", "Mongo3Interface");
 #endif   // JAVA_SUPPORT
 #else   // !MONGO_SUPPORT
-    sprintf(g->Message, MSG(NO_FEAT_SUPPORT), "MONGO");
+    snprintf(g->Message, sizeof(g->Message), MSG(NO_FEAT_SUPPORT), "MONGO");
     return true;
 #endif  // !MONGO_SUPPORT
   } // endif Uri
@@ -1201,7 +1253,7 @@ PTDB BSONDEF::GetTable(PGLOBAL g, MODE m)
 
     if (Lrecl) {
       // Allocate the parse work memory
-      G = PlugInit(NULL, (size_t)Lrecl * (Pretty < 0 ? 2 : 4));
+      G = PlugInit(NULL, (size_t)Lrecl * (Pretty < 0 ? 3 : 5));
     } else {
       strcpy(g->Message, "LRECL is not defined");
       return NULL;
@@ -1214,14 +1266,14 @@ PTDB BSONDEF::GetTable(PGLOBAL g, MODE m)
 #if defined(CMGO_SUPPORT)
         txfp = new(g) CMGFAM(this);
 #else
-        sprintf(g->Message, "Mongo %s Driver not available", "C");
+        snprintf(g->Message, sizeof(g->Message), "Mongo %s Driver not available", "C");
         return NULL;
 #endif
       } else if (Driver && toupper(*Driver) == 'J') {
 #if defined(JAVA_SUPPORT)
         txfp = new(g) JMGFAM(this);
 #else
-        sprintf(g->Message, "Mongo %s Driver not available", "Java");
+        snprintf(g->Message, sizeof(g->Message), "Mongo %s Driver not available", "Java");
         return NULL;
 #endif
       } else {             // Driver not specified
@@ -1230,11 +1282,12 @@ PTDB BSONDEF::GetTable(PGLOBAL g, MODE m)
 #elif defined(JAVA_SUPPORT)
         txfp = new(g) JMGFAM(this);
 #else   // !MONGO_SUPPORT
-        sprintf(g->Message, MSG(NO_FEAT_SUPPORT), "MONGO");
+        snprintf(g->Message, sizeof(g->Message), MSG(NO_FEAT_SUPPORT), "MONGO");
         return NULL;
 #endif  // !MONGO_SUPPORT
       } // endif Driver
 
+      Pretty = 4;   // Not a file
     } else if (Zipped) {
 #if defined(ZIP_SUPPORT)
       if (m == MODE_READ || m == MODE_ANY || m == MODE_ALTER) {
@@ -1246,7 +1299,7 @@ PTDB BSONDEF::GetTable(PGLOBAL g, MODE m)
         return NULL;
       } // endif's m
 #else   // !ZIP_SUPPORT
-      sprintf(g->Message, MSG(NO_FEAT_SUPPORT), "ZIP");
+      snprintf(g->Message, sizeof(g->Message), MSG(NO_FEAT_SUPPORT), "ZIP");
       return NULL;
 #endif  // !ZIP_SUPPORT
     } else if (Compressed) {
@@ -1256,7 +1309,7 @@ PTDB BSONDEF::GetTable(PGLOBAL g, MODE m)
       else
         txfp = new(g) ZLBFAM(this);
 #else   // !GZ_SUPPORT
-      sprintf(g->Message, MSG(NO_FEAT_SUPPORT), "GZ");
+      snprintf(g->Message, sizeof(g->Message), MSG(NO_FEAT_SUPPORT), "GZ");
       return NULL;
 #endif  // !GZ_SUPPORT
     } else if (map) {
@@ -1279,7 +1332,7 @@ PTDB BSONDEF::GetTable(PGLOBAL g, MODE m)
         return NULL;
       } // endif's m
 #else   // !ZIP_SUPPORT
-      sprintf(g->Message, MSG(NO_FEAT_SUPPORT), "ZIP");
+      snprintf(g->Message, sizeof(g->Message), MSG(NO_FEAT_SUPPORT), "ZIP");
       return NULL;
 #endif  // !ZIP_SUPPORT
     } else
@@ -1440,7 +1493,7 @@ int TDBBSN::EstimatedLength(void)
 } // end of Estimated Length
 
 /***********************************************************************/
-/*  OpenDB: Data Base open routine for JSN access method.              */
+/*  OpenDB: Data Base open routine for BSN access method.              */
 /***********************************************************************/
 bool TDBBSN::OpenDB(PGLOBAL g)
 {
@@ -1501,7 +1554,7 @@ bool TDBBSN::OpenDB(PGLOBAL g)
       case MODE_ARRAY:  type = TYPE_JAR;  break;
       case MODE_VALUE:  type = TYPE_JVAL; break;
       default:
-        sprintf(g->Message, "Invalid Jmode %d", Jmode);
+        snprintf(g->Message, sizeof(g->Message), "Invalid Jmode %d", Jmode);
         return true;
     } // endswitch Jmode
 
@@ -1609,7 +1662,7 @@ bool TDBBSN::PrepareWriting(PGLOBAL g)
 
       if ((signed)strlen(s) > Lrecl) {
         strncpy(To_Line, s, Lrecl);
-        sprintf(g->Message, "Line truncated (lrecl=%d)", Lrecl);
+        snprintf(g->Message, sizeof(g->Message), "Line truncated (lrecl=%d)", Lrecl);
         return PushWarning(g, this);
       } else
         strcpy(To_Line, s);
@@ -1662,6 +1715,7 @@ BSONCOL::BSONCOL(PGLOBAL g, PCOLDEF cdp, PTDB tdbp, PCOL cprec, int i)
   Xpd = false;
   Parsed = false;
   Warned = false;
+  Sgfy = false;
 } // end of BSONCOL constructor
 
 /***********************************************************************/
@@ -1681,6 +1735,7 @@ BSONCOL::BSONCOL(BSONCOL* col1, PTDB tdbp) : DOSCOL(col1, tdbp)
   Xpd = col1->Xpd;
   Parsed = col1->Parsed;
   Warned = col1->Warned;
+  Sgfy = col1->Sgfy;
 } // end of BSONCOL copy constructor
 
 /***********************************************************************/
@@ -1733,7 +1788,7 @@ bool BSONCOL::SetArrayOptions(PGLOBAL g, char* p, int i, PSZ nm)
       p[--n] = 0;
     } else if (!IsNum(p)) {
       // Wrong array specification
-      sprintf(g->Message, "Invalid array specification %s for %s", p, Name);
+      snprintf(g->Message, sizeof(g->Message), "Invalid array specification %s for %s", p, Name);
       return true;
     } // endif p
 
@@ -1792,7 +1847,7 @@ bool BSONCOL::SetArrayOptions(PGLOBAL g, char* p, int i, PSZ nm)
 
       break;
     default:
-      sprintf(g->Message,
+      snprintf(g->Message, sizeof(g->Message),
         "Invalid function specification %c for %s", *p, Name);
       return true;
     } // endswitch *p
@@ -1808,7 +1863,7 @@ bool BSONCOL::SetArrayOptions(PGLOBAL g, char* p, int i, PSZ nm)
     } // endif n
 
   } else {
-    sprintf(g->Message, "Wrong array specification for %s", Name);
+    snprintf(g->Message, sizeof(g->Message), "Wrong array specification for %s", Name);
     return true;
   } // endif's
 
@@ -1877,7 +1932,7 @@ bool BSONCOL::ParseJpath(PGLOBAL g)
         goto fin;
       } // endif Name
 
-    sprintf(g->Message, "Cannot parse updated column %s", Name);
+    snprintf(g->Message, sizeof(g->Message), "Cannot parse updated column %s", Name);
     return true;
   } // endif To_Orig
 
@@ -1919,6 +1974,10 @@ bool BSONCOL::ParseJpath(PGLOBAL g)
       // Analyse intermediate array processing
       if (SetArrayOptions(g, p, i, Nodes[i - 1].Key))
         return true;
+      else if (Xpd && Tbp->Mode == MODE_DELETE) {
+        strcpy(g->Message, "Cannot delete expanded columns");
+        return true;
+      } // endif Xpd
 
     } else if (*p == '*') {
       // Return JSON
@@ -1952,8 +2011,10 @@ PSZ BSONCOL::GetJpath(PGLOBAL g, bool proj)
       if (*p1 == '$') p1++;
       if (*p1 == '.') p1++;
       mgopath = PlugDup(g, p1);
-    } else
+    } else {
+      Sgfy = true;
       return NULL;
+    } // endif
 
     for (p1 = p2 = mgopath; *p1; p1++)
     {
@@ -1992,6 +2053,7 @@ PSZ BSONCOL::GetJpath(PGLOBAL g, bool proj)
       case '*':
         if (*(p2 - 1) == '.' && !*(p1 + 1)) {
           p2--;              // Suppress last :*
+          Sgfy = true;
           break;
         } // endif p2
         /* fall through */
@@ -2000,6 +2062,10 @@ PSZ BSONCOL::GetJpath(PGLOBAL g, bool proj)
         break;
       } // endswitch p1;
     }
+
+    if (*(p2 - 1) == '.')
+      p2--;
+
     *p2 = 0;
     return mgopath;
   } else
@@ -2108,7 +2174,7 @@ void BSONCOL::WriteColumn(PGLOBAL g)
 
     break;
   default:                  // ??????????
-    sprintf(g->Message, "Invalid column type %d", Buf_Type);
+    snprintf(g->Message, sizeof(g->Message), "Invalid column type %d", Buf_Type);
   } // endswitch Type
 
 } // end of WriteColumn
@@ -2121,7 +2187,9 @@ void BSONCOL::WriteColumn(PGLOBAL g)
 TDBBSON::TDBBSON(PGLOBAL g, PBDEF tdp, PTXF txfp) : TDBBSN(g, tdp, txfp)
 {
   Docp = NULL;
+  Docrow = NULL;
   Multiple = tdp->Multiple;
+  Docsize = 0;
   Done = Changed = false;
   Bp->SetPretty(2);
 } // end of TDBBSON standard constructor
@@ -2129,7 +2197,9 @@ TDBBSON::TDBBSON(PGLOBAL g, PBDEF tdp, PTXF txfp) : TDBBSN(g, tdp, txfp)
 TDBBSON::TDBBSON(PBTDB tdbp) : TDBBSN(tdbp)
 {
   Docp = tdbp->Docp;
+  Docrow = tdbp->Docrow;
   Multiple = tdbp->Multiple;
+  Docsize = tdbp->Docsize;
   Done = tdbp->Done;
   Changed = tdbp->Changed;
 } // end of TDBBSON copy constructor
@@ -2216,8 +2286,6 @@ int TDBBSON::MakeDocument(PGLOBAL g)
     return RC_FX;
 
   if ((objpath = PlugDup(g, Objname))) {
-    if (*objpath == '$') objpath++;
-    if (*objpath == '.') objpath++;
     p1 = (*objpath == '[') ? objpath++ : NULL;
 
     /*********************************************************************/
@@ -2256,7 +2324,7 @@ int TDBBSON::MakeDocument(PGLOBAL g)
         val = Bp->GetKeyValue(objp, key);
 
         if (!val || !(jsp = Bp->GetBson(val))) {
-          sprintf(g->Message, "Cannot find object key %s", key);
+          snprintf(g->Message, sizeof(g->Message), "Cannot find object key %s", key);
           return RC_FX;
         } // endif val
 
@@ -2264,7 +2332,7 @@ int TDBBSON::MakeDocument(PGLOBAL g)
         if (*p == '[') {
           // Old style
           if (p[strlen(p) - 1] != ']') {
-            sprintf(g->Message, "Invalid Table path near %s", p);
+            snprintf(g->Message, sizeof(g->Message), "Invalid Table path near %s", p);
             return RC_FX;
           } else
             p++;
@@ -2282,7 +2350,7 @@ int TDBBSON::MakeDocument(PGLOBAL g)
         val = Bp->GetArrayValue(arp, i);
 
         if (!val) {
-          sprintf(g->Message, "Cannot find array value %d", i);
+          snprintf(g->Message, sizeof(g->Message), "Cannot find array value %d", i);
           return RC_FX;
         } // endif val
 
@@ -2312,6 +2380,7 @@ int TDBBSON::MakeDocument(PGLOBAL g)
 
   } // endif jsp
 
+  Docsize = Bp->GetSize(Docp);
   Done = true;
   return RC_OK;
 } // end of MakeDocument
@@ -2326,7 +2395,7 @@ int TDBBSON::Cardinality(PGLOBAL g)
   else if (Cardinal < 0) {
     if (!Multiple) {
       if (MakeDocument(g) == RC_OK)
-        Cardinal = Bp->GetSize(Docp);
+        Cardinal = Docsize;
 
     } else
       return 10;
@@ -2355,6 +2424,7 @@ void TDBBSON::ResetSize(void)
   MaxSize = Cardinal = -1;
   Fpos = -1;
   N = 0;
+  Docrow = NULL;
   Done = false;
 } // end of ResetSize
 
@@ -2413,6 +2483,7 @@ bool TDBBSON::SetRecpos(PGLOBAL, int recpos)
 #endif // 0
 
   Fpos = recpos - 1;
+  Docrow = NULL;
   return false;
 } // end of SetRecpos
 
@@ -2428,6 +2499,7 @@ bool TDBBSON::OpenDB(PGLOBAL g)
     Fpos = -1;
     NextSame = false;
     SameRow = 0;
+    Docrow = NULL;
     return false;
   } // endif use
 
@@ -2443,7 +2515,7 @@ bool TDBBSON::OpenDB(PGLOBAL g)
     case MODE_ARRAY:  Row = Bp->NewVal(TYPE_JAR);  break;
     case MODE_VALUE:  Row = Bp->NewVal(TYPE_JVAL); break;
     default:
-      sprintf(g->Message, "Invalid Jmode %d", Jmode);
+      snprintf(g->Message, sizeof(g->Message), "Invalid Jmode %d", Jmode);
       return true;
     } // endswitch Jmode
 
@@ -2468,12 +2540,9 @@ int TDBBSON::ReadDB(PGLOBAL)
     NextSame = false;
     M++;
     rc = RC_OK;
-  } else if (++Fpos < (signed)Bp->GetSize(Docp)) {
-    Row = Bp->GetArrayValue(Docp, Fpos);
-
-  if (Row->Type == TYPE_JVAL)
-      Row = Bp->GetBson(Row);
-
+  } else if (++Fpos < Docsize) {
+    Docrow = (Docrow) ? Bp->GetNext(Docrow) : Bp->GetArrayValue(Docp, Fpos);
+    Row = (Docrow->Type == TYPE_JVAL) ? Bp->GetBson(Docrow) : Docrow;
     SameRow = 0;
     M = 1;
     rc = RC_OK;
